@@ -28,7 +28,6 @@
 #include "print-tree.h"
 #include "transaction.h"
 #include "list.h"
-#include "version.h"
 #include "utils.h"
 
 #define FIELD_BUF_LEN 80
@@ -110,6 +109,7 @@ static void print_usage(void)
 		"to corrupt and a root+key for the item)\n");
 	fprintf(stderr, "\t-D Corrupt a dir item, must specify key and field\n");
 	fprintf(stderr, "\t-d Delete this item (must specify -K)\n");
+	fprintf(stderr, "\t-r Operate on this root (only works with -d)\n");
 	exit(1);
 }
 
@@ -158,7 +158,7 @@ static int corrupt_keys_in_block(struct btrfs_root *root, u64 bytenr)
 	struct extent_buffer *eb;
 
 	eb = read_tree_block(root, bytenr, root->leafsize, 0);
-	if (!eb)
+	if (!extent_buffer_uptodate(eb))
 		return -EIO;;
 
 	corrupt_keys(NULL, root, eb);
@@ -286,7 +286,7 @@ static void btrfs_corrupt_extent_tree(struct btrfs_trans_handle *trans,
 		next = read_tree_block(root, btrfs_node_blockptr(eb, i),
 				       root->leafsize,
 				       btrfs_node_ptr_generation(eb, i));
-		if (!next)
+		if (!extent_buffer_uptodate(next))
 			continue;
 		btrfs_corrupt_extent_tree(trans, root, next);
 		free_extent_buffer(next);
@@ -694,7 +694,7 @@ static int corrupt_metadata_block(struct btrfs_root *root, u64 block,
 	}
 
 	eb = read_tree_block(root, block, root->leafsize, 0);
-	if (!eb) {
+	if (!extent_buffer_uptodate(eb)) {
 		fprintf(stderr, "Couldn't read in tree block %s\n", field);
 		return -EINVAL;
 	}
@@ -1008,6 +1008,7 @@ int main(int ac, char **av)
 	u64 metadata_block = 0;
 	u64 inode = 0;
 	u64 file_extent = (u64)-1;
+	u64 root_objectid = 0;
 	char field[FIELD_BUF_LEN];
 
 	field[0] = '\0';
@@ -1016,30 +1017,30 @@ int main(int ac, char **av)
 
 	while(1) {
 		int c;
-		int option_index = 0;
 		static const struct option long_options[] = {
 			/* { "byte-count", 1, NULL, 'b' }, */
-			{ "logical", 1, NULL, 'l' },
-			{ "copy", 1, NULL, 'c' },
-			{ "bytes", 1, NULL, 'b' },
-			{ "extent-record", 0, NULL, 'e' },
-			{ "extent-tree", 0, NULL, 'E' },
-			{ "keys", 0, NULL, 'k' },
-			{ "chunk-record", 0, NULL, 'u' },
-			{ "chunk-tree", 0, NULL, 'U' },
-			{ "inode", 1, NULL, 'i'},
-			{ "file-extent", 1, NULL, 'x'},
-			{ "metadata-block", 1, NULL, 'm'},
-			{ "field", 1, NULL, 'f'},
-			{ "key", 1, NULL, 'K'},
-			{ "item", 0, NULL, 'I'},
-			{ "dir-item", 0, NULL, 'D'},
-			{ "delete", 0, NULL, 'd'},
+			{ "logical", required_argument, NULL, 'l' },
+			{ "copy", required_argument, NULL, 'c' },
+			{ "bytes", required_argument, NULL, 'b' },
+			{ "extent-record", no_argument, NULL, 'e' },
+			{ "extent-tree", no_argument, NULL, 'E' },
+			{ "keys", no_argument, NULL, 'k' },
+			{ "chunk-record", no_argument, NULL, 'u' },
+			{ "chunk-tree", no_argument, NULL, 'U' },
+			{ "inode", required_argument, NULL, 'i'},
+			{ "file-extent", required_argument, NULL, 'x'},
+			{ "metadata-block", required_argument, NULL, 'm'},
+			{ "field", required_argument, NULL, 'f'},
+			{ "key", required_argument, NULL, 'K'},
+			{ "item", no_argument, NULL, 'I'},
+			{ "dir-item", no_argument, NULL, 'D'},
+			{ "delete", no_argument, NULL, 'd'},
+			{ "root", no_argument, NULL, 'r'},
 			{ NULL, 0, NULL, 0 }
 		};
 
-		c = getopt_long(ac, av, "l:c:b:eEkuUi:f:x:m:K:IDd", long_options,
-				&option_index);
+		c = getopt_long(ac, av, "l:c:b:eEkuUi:f:x:m:K:IDdr:",
+				long_options, NULL);
 		if (c < 0)
 			break;
 		switch(c) {
@@ -1099,6 +1100,9 @@ int main(int ac, char **av)
 			case 'd':
 				delete = 1;
 				break;
+			case 'r':
+				root_objectid = arg_strtou64(optarg);
+				break;
 			default:
 				print_usage();
 		}
@@ -1207,9 +1211,25 @@ int main(int ac, char **av)
 		ret = corrupt_btrfs_item(root, &key, field);
 	}
 	if (delete) {
+		struct btrfs_root *target = root;
+
 		if (!key.objectid)
 			print_usage();
-		ret = delete_item(root, &key);
+		if (root_objectid) {
+			struct btrfs_key root_key;
+
+			root_key.objectid = root_objectid;
+			root_key.type = BTRFS_ROOT_ITEM_KEY;
+			root_key.offset = (u64)-1;
+
+			target = btrfs_read_fs_root(root->fs_info, &root_key);
+			if (IS_ERR(target)) {
+				fprintf(stderr, "Couldn't find root %llu\n",
+					(unsigned long long)root_objectid);
+				print_usage();
+			}
+		}
+		ret = delete_item(target, &key);
 		goto out_close;
 	}
 	if (key.objectid || key.offset || key.type) {
